@@ -18,6 +18,19 @@ static const t_protocolInfo protocolInfos[] = {
 	{ UNKNOWN, _T(""),      21,  false, _T("") }
 };
 
+static const wxString typeNames[SERVERTYPE_MAX] = {
+	wxTRANSLATE("Default (Autodetect)"),
+	_T("Unix"),
+	_T("VMS"),
+	_T("DOS"),
+	_T("MVS, OS/390, z/OS"),
+	_T("VxWorks"),
+	_T("z/VM"),
+	_T("HP NonStop"),
+	wxTRANSLATE("DOS-like with virtual paths"),
+	_T("Cygwin")
+};
+
 CServer::CServer()
 {
 	Initialize();
@@ -99,7 +112,7 @@ bool CServer::ParseUrl(wxString host, unsigned int port, wxString user, wxString
 		user.Trim(true);
 		user.Trim(false);
 
-		if (user == _T(""))
+		if (user == _T("") && m_logonType != ASK && m_logonType != INTERACTIVE)
 		{
 			user = _T("anonymous");
 			pass = _T("anonymous@example.com");
@@ -113,7 +126,29 @@ bool CServer::ParseUrl(wxString host, unsigned int port, wxString user, wxString
 		host = host.Left(pos);
 	}
 
-	pos = host.Find(':');
+	if (host[0] == '[')
+	{
+		// Probably IPv6 address
+		pos = host.Find(']');
+		if (pos == -1)
+		{
+			error = _("Host starts with '[' but no closing bracket found.");
+			return false;
+		}
+		if (host[pos + 1])
+		{
+			if (host[pos + 1] != ':')
+			{
+				error = _("Invalid host, after closing bracket only colon and port may follow.");
+				return false;
+			}
+			pos++;
+		}
+		else
+			pos = -1;
+	}
+	else
+		pos = host.Find(':');
 	if (pos != -1)
 	{
 		if (!pos)
@@ -129,7 +164,6 @@ bool CServer::ParseUrl(wxString host, unsigned int port, wxString user, wxString
 			return false;
 		}
 		port = tmp;
-
 		host = host.Left(pos);
 	}
 	else
@@ -153,14 +187,24 @@ bool CServer::ParseUrl(wxString host, unsigned int port, wxString user, wxString
 	}
 
 	m_host = host;
+
+	if (m_host[0] == '[')
+	{
+		m_host.RemoveLast();
+		m_host = m_host.Mid(1);
+	}
+
 	m_port = port;
 	m_user = user;
 	m_pass = pass;
 	m_account = _T("");
-	if (m_user == _T("") || m_user == _T("anonymous"))
-		m_logonType = ANONYMOUS;
-	else
-		m_logonType = NORMAL;
+	if (m_logonType != ASK && m_logonType != INTERACTIVE)
+	{
+		if (m_user == _T("") || m_user == _T("anonymous"))
+			m_logonType = ANONYMOUS;
+		else
+			m_logonType = NORMAL;
+	}
 
 	if (m_protocol == UNKNOWN)
 		m_protocol = GetProtocolFromPort(port);
@@ -192,7 +236,7 @@ wxString CServer::GetUser() const
 {
 	if (m_logonType == ANONYMOUS)
 		return _T("anonymous");
-	
+
 	return m_user;
 }
 
@@ -200,7 +244,7 @@ wxString CServer::GetPass() const
 {
 	if (m_logonType == ANONYMOUS)
 		return _T("anon@localhost");
-	
+
 	return m_pass;
 }
 
@@ -228,6 +272,8 @@ CServer& CServer::operator=(const CServer &op)
 	m_encodingType = op.m_encodingType;
 	m_customEncoding = op.m_customEncoding;
 	m_postLoginCommands = op.m_postLoginCommands;
+	m_bypassProxy = op.m_bypassProxy;
+	m_name = op.m_name;
 
 	return *this;
 }
@@ -275,6 +321,8 @@ bool CServer::operator==(const CServer &op) const
 	}
 	if (m_postLoginCommands != op.m_postLoginCommands)
 		return false;
+	if (m_bypassProxy != op.m_bypassProxy)
+		return false;
 
 	// Do not compare number of allowed multiple connections
 
@@ -344,7 +392,7 @@ bool CServer::operator<(const CServer &op) const
 		return true;
 	else if (m_timezoneOffset > op.m_timezoneOffset)
 		return false;
-	
+
 	if (m_pasvMode < op.m_pasvMode)
 		return true;
 	else if (m_pasvMode > op.m_pasvMode)
@@ -362,6 +410,10 @@ bool CServer::operator<(const CServer &op) const
 		else if (m_customEncoding > op.m_customEncoding)
 			return false;
 	}
+	if (m_bypassProxy < op.m_bypassProxy)
+		return true;
+	else if (m_bypassProxy > op.m_bypassProxy)
+		return false;
 
 	// Do not compare number of allowed multiple connections
 
@@ -373,6 +425,44 @@ bool CServer::operator!=(const CServer &op) const
 	return !(*this == op);
 }
 
+bool CServer::EqualsNoPass(const CServer &op) const
+{
+	if (m_protocol != op.m_protocol)
+		return false;
+	else if (m_type != op.m_type)
+		return false;
+	else if (m_host != op.m_host)
+		return false;
+	else if (m_port != op.m_port)
+		return false;
+	else if (m_logonType != op.m_logonType)
+		return false;
+	else if (m_logonType != ANONYMOUS)
+	{
+		if (m_user != op.m_user)
+			return false;
+	}
+	if (m_timezoneOffset != op.m_timezoneOffset)
+		return false;
+	else if (m_pasvMode != op.m_pasvMode)
+		return false;
+	else if (m_encodingType != op.m_encodingType)
+		return false;
+	else if (m_encodingType == ENCODING_CUSTOM)
+	{
+		if (m_customEncoding != op.m_customEncoding)
+			return false;
+	}
+	if (m_postLoginCommands != op.m_postLoginCommands)
+		return false;
+	if (m_bypassProxy != op.m_bypassProxy)
+		return false;
+
+	// Do not compare number of allowed multiple connections
+
+	return true;
+}
+
 CServer::CServer(enum ServerProtocol protocol, enum ServerType type, wxString host, unsigned int port, wxString user, wxString pass /*=_T("")*/, wxString account /*=_T("")*/)
 {
 	Initialize();
@@ -407,6 +497,7 @@ enum LogonType CServer::GetLogonType() const
 
 void CServer::SetLogonType(enum LogonType logonType)
 {
+	wxASSERT(logonType != LOGONTYPE_MAX);
 	m_logonType = logonType;
 }
 
@@ -433,7 +524,7 @@ bool CServer::SetHost(wxString host, unsigned int port)
 
 	if (m_protocol == UNKNOWN)
 		m_protocol = GetProtocolFromPort(m_port);
-	
+
 	return true;
 }
 
@@ -443,10 +534,15 @@ bool CServer::SetUser(const wxString& user, const wxString& pass /*=_T("")*/)
 		return true;
 
 	if (user == _T(""))
-		return false;
-
+	{
+		if (m_logonType != ASK && m_logonType != INTERACTIVE)
+			return false;
+		m_pass = _T("");
+	}
+	else
+		m_pass = pass;
+	
 	m_user = user;
-	m_pass = pass;
 	
 	return true;
 }
@@ -496,17 +592,23 @@ int CServer::MaximumMultipleConnections() const
 	return m_maximumMultipleConnections;
 }
 
-wxString CServer::FormatHost() const
+wxString CServer::FormatHost(bool always_omit_port /*=false*/) const
 {
 	wxString host = m_host;
 
-	if (m_port != GetDefaultPort(m_protocol))
-		host += wxString::Format(_T(":%d"), m_port);
+	if (host.Find(':') != -1)
+		host = _T("[") + host + _T("]");
+
+	if (!always_omit_port)
+	{
+		if (m_port != GetDefaultPort(m_protocol))
+			host += wxString::Format(_T(":%d"), m_port);
+	}
 
 	return host;
 }
 
-wxString CServer::FormatServer() const
+wxString CServer::FormatServer(const bool always_include_prefix /*=false*/) const
 {
 	wxString server = FormatHost();
 
@@ -520,10 +622,13 @@ wxString CServer::FormatServer() const
 			wxString prefix = GetPrefixFromProtocol(m_protocol);
 			if (prefix != _T(""))
 				server = prefix + _T("://") + server;
+			else if (always_include_prefix)
+				server = prefix + _T("://") + server;
 		}
 		break;
 	case FTP:
-		if (GetProtocolFromPort(m_port) != FTP && GetProtocolFromPort(m_port) != UNKNOWN)
+		if (always_include_prefix ||
+			(GetProtocolFromPort(m_port) != FTP && GetProtocolFromPort(m_port) != UNKNOWN))
 			server = _T("ftp://") + server;
 		break;
 	}
@@ -546,6 +651,7 @@ void CServer::Initialize()
 	m_maximumMultipleConnections = 0;
 	m_encodingType = ENCODING_AUTO;
 	m_customEncoding = _T("");
+	m_bypassProxy = false;
 }
 
 bool CServer::SetEncodingType(enum CharsetEncoding type, const wxString& encoding /*=_T("")*/)
@@ -566,7 +672,7 @@ bool CServer::SetCustomEncoding(const wxString& encoding)
 
 	m_encodingType = ENCODING_CUSTOM;
 	m_customEncoding = encoding;
-	
+
 	return true;
 }
 
@@ -592,15 +698,18 @@ unsigned int CServer::GetDefaultPort(enum ServerProtocol protocol)
 	return 21;
 }
 
-enum ServerProtocol CServer::GetProtocolFromPort(unsigned int port)
+enum ServerProtocol CServer::GetProtocolFromPort(unsigned int port, bool defaultOnly /*=false*/)
 {
 	for (unsigned int i = 0; protocolInfos[i].protocol != UNKNOWN; i++)
 	{
 		if (protocolInfos[i].defaultPort == port)
 			return protocolInfos[i].protocol;
 	}
-	
-	// Default to FTP
+
+	if (defaultOnly)
+		return UNKNOWN;
+
+	// Else default to FTP
 	return FTP;
 }
 
@@ -678,3 +787,72 @@ wxString CServer::GetPrefixFromProtocol(const enum ServerProtocol protocol)
 
 	return _T("");
 }
+
+void CServer::SetBypassProxy(bool val)
+{
+  m_bypassProxy = val;
+}
+
+bool CServer::GetBypassProxy() const
+{
+  return m_bypassProxy;
+}
+
+bool CServer::ProtocolHasDataTypeConcept(const enum ServerProtocol protocol)
+{
+	if (protocol == FTP || protocol == FTPES || protocol == FTPES)
+		return true;
+
+	return false;
+}
+
+wxString CServer::GetNameFromServerType(enum ServerType type)
+{
+	wxASSERT(type != SERVERTYPE_MAX);
+	return wxGetTranslation(typeNames[type]);
+}
+
+enum ServerType CServer::GetServerTypeFromName(const wxString& name)
+{
+	for (int i = 0; i < SERVERTYPE_MAX; i++)
+	{
+		enum ServerType type = (enum ServerType)i;
+		if (name == CServer::GetNameFromServerType(type))
+			return type;
+	}
+
+	return DEFAULT;
+}
+
+enum LogonType CServer::GetLogonTypeFromName(const wxString& name)
+{
+	if (name == _("Normal"))
+		return NORMAL;
+	else if (name == _("Ask for password"))
+		return ASK;
+	else if (name == _("Interactive"))
+		return INTERACTIVE;
+	else if (name == _("Account"))
+		return ACCOUNT;
+	else
+		return ANONYMOUS;
+}
+
+wxString CServer::GetNameFromLogonType(enum LogonType type)
+{
+	wxASSERT(type != LOGONTYPE_MAX);
+
+	switch (type)
+	{
+	case NORMAL:
+		return _("Normal");
+	case ASK:
+		return _("Ask for password");
+	case INTERACTIVE:
+		return _("Interactive");
+	case ACCOUNT:
+		return _("Account");
+	default:
+		return _("Anonymous");
+	}
+}
