@@ -5,6 +5,9 @@
 #include "filter.h"
 #include <wx/dnd.h>
 #include "dndobjects.h"
+#include "inputdialog.h"
+#include "local_filesys.h"
+#include "dragdropmanager.h"
 
 #ifdef _DEBUG
 #define new DEBUG_NEW
@@ -13,8 +16,17 @@
 #ifdef __WXMSW__
 #include <wx/msw/registry.h>
 #include <shlobj.h>
+#include <dbt.h>
+#include "volume_enumerator.h"
 #endif
 
+class CTreeItemData : public wxTreeItemData
+{
+public:
+	CTreeItemData(const wxString& known_subdir) : m_known_subdir(known_subdir) {}
+	wxString m_known_subdir;
+};
+
 class CLocalTreeViewDropTarget : public wxDropTarget
 {
 public:
@@ -72,15 +84,19 @@ public:
 		if (!hit)
 			return wxDragNone;
 
-		wxString dir = GetDirFromItem(hit);
-		if (dir == _T("") || !CState::LocalDirIsWriteable(dir))
+		const CLocalPath path(GetDirFromItem(hit));
+		if (path.empty() || !path.IsWriteable())
 			return wxDragNone;
 
 		if (!GetData())
 			return wxDragError;
 
+		CDragDropManager* pDragDropManager = CDragDropManager::Get();
+		if (pDragDropManager)
+			pDragDropManager->pDropTarget = m_pLocalTreeView;
+
 		if (m_pDataObject->GetReceivedFormat() == m_pFileDataObject->GetFormat())
-			m_pLocalTreeView->m_pState->HandleDroppedFiles(m_pFileDataObject, dir, def == wxDragCopy);
+			m_pLocalTreeView->m_pState->HandleDroppedFiles(m_pFileDataObject, path, def == wxDragCopy);
 		else
 		{
 			if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId())
@@ -95,7 +111,7 @@ public:
 				return wxDragNone;
 			}
 
-			if (!m_pLocalTreeView->m_pState->DownloadDroppedFiles(m_pRemoteDataObject, dir))
+			if (!m_pLocalTreeView->m_pState->DownloadDroppedFiles(m_pRemoteDataObject, path))
 				return wxDragNone;
 		}
 
@@ -110,8 +126,8 @@ public:
 		if (!hit)
 			return false;
 
-		wxString dir = GetDirFromItem(hit);
-		if (dir == _T("") || !CState::LocalDirIsWriteable(dir))
+		const wxString dir = GetDirFromItem(hit);
+		if (dir == _T("") || !CLocalPath(dir).IsWriteable())
 			return false;
 
 		return true;
@@ -192,15 +208,32 @@ EVT_TREE_BEGIN_DRAG(wxID_ANY, CLocalTreeView::OnBeginDrag)
 EVT_KEY_DOWN(CLocalTreeView::OnKeyDown)
 #else
 EVT_COMMAND(-1, fzEVT_VOLUMESENUMERATED, CLocalTreeView::OnVolumesEnumerated)
+EVT_COMMAND(-1, fzEVT_VOLUMEENUMERATED, CLocalTreeView::OnVolumesEnumerated)
 #endif //__WXMSW__
+EVT_TREE_ITEM_MENU(wxID_ANY, CLocalTreeView::OnContextMenu)
+EVT_MENU(XRCID("ID_UPLOAD"), CLocalTreeView::OnMenuUpload)
+EVT_MENU(XRCID("ID_ADDTOQUEUE"), CLocalTreeView::OnMenuUpload)
+EVT_MENU(XRCID("ID_DELETE"), CLocalTreeView::OnMenuDelete)
+EVT_MENU(XRCID("ID_RENAME"), CLocalTreeView::OnMenuRename)
+EVT_MENU(XRCID("ID_MKDIR"), CLocalTreeView::OnMenuMkdir)
+EVT_TREE_BEGIN_LABEL_EDIT(wxID_ANY, CLocalTreeView::OnBeginLabelEdit)
+EVT_TREE_END_LABEL_EDIT(wxID_ANY, CLocalTreeView::OnEndLabelEdit)
+EVT_CHAR(CLocalTreeView::OnChar)
 END_EVENT_TABLE()
 
 CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState *pState, CQueueView *pQueueView)
 	: wxTreeCtrl(parent, id, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxTR_LINES_AT_ROOT | wxTR_HAS_BUTTONS | wxNO_BORDER),
 	CSystemImageList(16),
-	CStateEventHandler(pState, STATECHANGE_LOCAL_DIR | STATECHANGE_APPLYFILTER),
+	CStateEventHandler(pState),
 	m_pQueueView(pQueueView)
 {
+#ifdef __WXMAC__
+	SetFont(wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT));
+#endif
+
+	pState->RegisterHandler(this, STATECHANGE_LOCAL_DIR);
+	pState->RegisterHandler(this, STATECHANGE_APPLYFILTER);
+
 	m_setSelection = false;
 
 	SetImageList(GetSystemImageList());
@@ -309,8 +342,11 @@ wxTreeItemId CLocalTreeView::GetNearestParent(wxString& localDir)
 		root = GetNextSibling(root);
 	}
 	if (!root)
-		// TODO: Number is drives changed
+	{
+		if (drive[1] == ':')
+			return AddDrive(drive[0]);
 		return wxTreeItemId();
+	}
 #else
 		if (localDir[0] == '/')
 			localDir = localDir.Mid(1);
@@ -396,14 +432,14 @@ bool CLocalTreeView::DisplayDrives(wxTreeItemId parent)
 		// Check if drive should be hidden by default
 		if (pDrive[0] != 0 && pDrive[1] == ':')
 		{
-			int bit = -1;
+			int bit = 0;
 			char letter = pDrive[0];
 			if (letter >= 'A' && letter <= 'Z')
 				bit = 1 << (letter - 'A');
 			if (letter >= 'a' && letter <= 'z')
 				bit = 1 << (letter - 'a');
 
-			if (bit != -1 && drivesToHide & bit)
+			if (drivesToHide & bit)
 			{
 				pDrive += wxStrlen(pDrive) + 1;
 				continue;
@@ -428,19 +464,51 @@ bool CLocalTreeView::DisplayDrives(wxTreeItemId parent)
 
 #endif
 
-void CLocalTreeView::DisplayDir(wxTreeItemId parent, const wxString& dirname, const wxString& filterException /*=_T("")*/)
+void CLocalTreeView::DisplayDir(wxTreeItemId parent, const wxString& dirname, const wxString& knownSubdir /*=_T("")*/)
 {
+	CLocalFileSystem local_filesystem;
+
+	{
+		wxLogNull log;
+		if (!local_filesystem.BeginFindFiles(dirname, true))
+		{
+			if (knownSubdir != _T(""))
+			{
+				wxTreeItemId item = GetSubdir(parent, knownSubdir);
+				if (item != wxTreeItemId())
+					return;
+
+				const wxString fullName = dirname + knownSubdir;
+				item = AppendItem(parent, knownSubdir, GetIconIndex(::dir, fullName), GetIconIndex(opened_dir, fullName));
+				CheckSubdirStatus(item, fullName);
+			}
+			else
+			{
+				m_setSelection = true;
+				DeleteChildren(parent);
+				m_setSelection = false;
+			}
+			return;
+		}
+	}
+
 	wxASSERT(parent);
 	m_setSelection = true;
 	DeleteChildren(parent);
 	m_setSelection = false;
-	wxDir dir(dirname);
+
+	CFilterManager filter;
+
+	bool matchedKnown = false;
+
 	wxString file;
-
-	CFilterDialog filter;
-
-	for (bool found = dir.GetFirst(&file, _T(""), wxDIR_DIRS | wxDIR_HIDDEN); found; found = dir.GetNext(&file))
+	bool wasLink;
+	int attributes;
+	bool is_dir;
+	const wxLongLong size(-1);
+	while (local_filesystem.GetNextFile(file, wasLink, is_dir, 0, 0, &attributes))
 	{
+		wxASSERT(is_dir);
 		if (file == _T(""))
 		{
 			wxGetApp().DisplayEncodingWarning();
@@ -449,49 +517,62 @@ void CLocalTreeView::DisplayDir(wxTreeItemId parent, const wxString& dirname, co
 
 		wxString fullName = dirname + file;
 #ifdef __WXMSW__
-		if (file.CmpNoCase(filterException))
+		if (file.CmpNoCase(knownSubdir))
 #else
-		if (file != filterException)
+		if (file != knownSubdir)
 #endif
 		{
-			wxFileName fn(fullName);
-			const bool isDir = fn.DirExists();
-
-			wxLongLong size;
-
-			wxStructStat buf;
-			if (!isDir && !wxStat(fn.GetFullPath(), &buf))
-				size = buf.st_size;
-			else
-				size = -1;
-
-			if (filter.FilenameFiltered(file, isDir, size, true))
+			if (filter.FilenameFiltered(file, dirname, true, size, true, attributes))
 				continue;
 		}
+		else
+			matchedKnown = true;
 
 		wxTreeItemId item = AppendItem(parent, file, GetIconIndex(::dir, fullName), GetIconIndex(opened_dir, fullName));
-		if (HasSubdir(fullName))
-			AppendItem(item, _T(""));
+		CheckSubdirStatus(item, fullName);
 	}
+
+	if (!matchedKnown && knownSubdir != _T(""))
+	{
+		const wxString fullName = dirname + knownSubdir;
+		wxTreeItemId item = AppendItem(parent, knownSubdir, GetIconIndex(::dir, fullName), GetIconIndex(opened_dir, fullName));
+		CheckSubdirStatus(item, fullName);
+	}
+
 	SortChildren(parent);
 }
 
-bool CLocalTreeView::HasSubdir(const wxString& dirname)
+wxString CLocalTreeView::HasSubdir(const wxString& dirname)
 {
 	wxLogNull nullLog;
-	wxDir dir(dirname);
-	if (!dir.IsOpened())
-		return false;
 
-	CFilterDialog filter;
+	CFilterManager filter;
+	
+	CLocalFileSystem local_filesystem;
+	if (!local_filesystem.BeginFindFiles(dirname, true))
+		return _T("");
+
 	wxString file;
-	for (bool found = dir.GetFirst(&file, _T(""), wxDIR_DIRS | wxDIR_HIDDEN); found; found = dir.GetNext(&file))
+	bool wasLink;
+	int attributes;
+	bool is_dir;
+	const wxLongLong size(-1);
+	while (local_filesystem.GetNextFile(file, wasLink, is_dir, 0, 0, &attributes))
 	{
-		if (!filter.FilenameFiltered(file, true, -1, true))
-			return true;
+		wxASSERT(is_dir);
+		if (file == _T(""))
+		{
+			wxGetApp().DisplayEncodingWarning();
+			continue;
+		}
+
+		if (filter.FilenameFiltered(file, dirname, true, size, true, attributes))
+			continue;
+
+		return file;
 	}
 
-	return false;
+	return _T("");
 }
 
 wxTreeItemId CLocalTreeView::MakeSubdirs(wxTreeItemId parent, wxString dirname, wxString subDir)
@@ -649,41 +730,54 @@ void CLocalTreeView::Refresh()
 	dirsToCheck.push_back(dir);
 #endif
 
-	CFilterDialog filter;
+	CFilterManager filter;
 
 	while (!dirsToCheck.empty())
 	{
 		t_dir dir = dirsToCheck.front();
 		dirsToCheck.pop_front();
 
-		wxDir find(dir.dir);
-		wxString file;
-		bool found = find.GetFirst(&file, _T(""), wxDIR_DIRS | wxDIR_HIDDEN);
+		CLocalFileSystem local_filesystem;
+		if (!local_filesystem.BeginFindFiles(dir.dir, true))
+		{
+			// Dir does exist (listed in parent) but may not be accessible.
+			// Recurse into children anyhow, they might be accessible again.
+			wxTreeItemIdValue value;
+			wxTreeItemId child = GetFirstChild(dir.item, value);
+			while (child)
+			{
+				t_dir subdir;
+				subdir.dir = dir.dir + GetItemText(child) + separator;
+				subdir.item = child;
+				dirsToCheck.push_back(subdir);
+
+				child = GetNextSibling(child);
+			}
+			continue;
+		}
+
 		std::list<wxString> dirs;
-		while (found)
+
+		
+		wxString file;
+		const wxLongLong size(-1);
+		bool was_link;
+		bool is_dir;
+		int attributes;
+		while (local_filesystem.GetNextFile(file, was_link, is_dir, 0, 0, &attributes))
 		{
 			if (file == _T(""))
 			{
 				wxGetApp().DisplayEncodingWarning();
-				found = find.GetNext(&file);
 				continue;
 			}
-			wxFileName fn(dir.dir + file);
-			const bool isDir = fn.DirExists();
 
-			wxLongLong size;
+			if (filter.FilenameFiltered(file, dir.dir, true, size, true, attributes))
+				continue;
 
-			wxStructStat buf;
-			if (!isDir && !wxStat(fn.GetFullPath(), &buf))
-				size = buf.st_size;
-			else
-				size = -1;
-
-			if (!filter.FilenameFiltered(file, isDir, size , true))
-				dirs.push_back(file);
-			found = find.GetNext(&file);
+			dirs.push_back(file);
 		}
-		dirs.sort(sortfunc);
+		dirs.sort(sortfunc);		
 
 		bool inserted = false;
 
@@ -700,21 +794,11 @@ void CLocalTreeView::Refresh()
 			{
 				if (!IsExpanded(child))
 				{
-					wxTreeItemId subchild = GetLastChild(child);
-					if (subchild && GetItemText(subchild) == _T(""))
-					{
-						if (!HasSubdir(dir.dir + *iter + separator))
-							Delete(subchild);
-					}
-					else if (!subchild)
-					{
-						if (HasSubdir(dir.dir + *iter + separator))
-							AppendItem(child, _T(""));
-					}
-					else
+					wxString path = dir.dir + *iter + separator;
+					if (!CheckSubdirStatus(child, path))
 					{
 						t_dir subdir;
-						subdir.dir = dir.dir + *iter + separator;
+						subdir.dir = path;
 						subdir.item = child;
 						dirsToCheck.push_front(subdir);
 					}
@@ -731,32 +815,38 @@ void CLocalTreeView::Refresh()
 			}
 			else if (cmp > 0)
 			{
+				wxTreeItemId sel = GetSelection();
+				while (sel && sel != child)
+					sel = GetItemParent(sel);
 				wxTreeItemId prev = GetPrevSibling(child);
-				Delete(child);
+				if (!sel)
+					Delete(child);
 				child = prev;
 			}
 			else if (cmp < 0)
 			{
 				wxString fullname = dir.dir + *iter + separator;
 				wxTreeItemId newItem = AppendItem(dir.item, *iter, GetIconIndex(::dir, fullname), GetIconIndex(opened_dir, fullname));
-				if (HasSubdir(fullname))
-					AppendItem(newItem, _T(""));
+				CheckSubdirStatus(newItem, fullname);
 				iter++;
 				inserted = true;
 			}
 		}
 		while (child)
 		{
+			wxTreeItemId sel = GetSelection();
+			while (sel && sel != child)
+				sel = GetItemParent(sel);
 			wxTreeItemId prev = GetPrevSibling(child);
-			Delete(child);
+			if (!sel)
+				Delete(child);
 			child = prev;
 		}
 		while (iter != dirs.rend())
 		{
 			wxString fullname = dir.dir + *iter + separator;
 			wxTreeItemId newItem = AppendItem(dir.item, *iter, GetIconIndex(::dir, fullname), GetIconIndex(opened_dir, fullname));
-			if (HasSubdir(fullname))
-				AppendItem(newItem, _T(""));
+			CheckSubdirStatus(newItem, fullname);
 			iter++;
 			inserted = true;
 		}
@@ -807,12 +897,15 @@ int CLocalTreeView::OnCompareItems(const wxTreeItemId& item1, const wxTreeItemId
 #endif
 }
 
-void CLocalTreeView::OnStateChange(unsigned int event, const wxString& data)
+void CLocalTreeView::OnStateChange(enum t_statechange_notifications notification, const wxString& data)
 {
-	if (event == STATECHANGE_LOCAL_DIR)
-		SetDir(m_pState->GetLocalDir());
-	else if (event == STATECHANGE_APPLYFILTER)
+	if (notification == STATECHANGE_LOCAL_DIR)
+		SetDir(m_pState->GetLocalDir().GetPath());
+	else
+	{
+		wxASSERT(notification == STATECHANGE_APPLYFILTER);
 		Refresh();
+	}
 }
 
 void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
@@ -830,14 +923,21 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 	if (dir == _T("/"))
 		return;
 
+#ifdef __WXMSW__
 	if (dir.Last() == '\\')
 		dir.RemoveLast();
+#endif
+	if (dir.Last() == '/')
+		dir.RemoveLast();
 
 #ifdef __WXMSW__
 	if (dir.Last() == ':')
 		return;
 #endif
 
+	CDragDropManager* pDragDropManager = CDragDropManager::Init();
+	pDragDropManager->pDragSource = this;
+
 	wxFileDataObject obj;
 
 	obj.AddFile(dir);
@@ -845,8 +945,17 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 	wxDropSource source(this);
 	source.SetData(obj);
 	int res = source.DoDragDrop(wxDrag_AllowMove);
-	if (res == wxDragCopy || res == wxDragMove)
+
+	bool handled_internally = pDragDropManager->pDropTarget != 0;
+
+	pDragDropManager->Release();
+
+	if (!handled_internally && (res == wxDragCopy || res == wxDragMove))
+	{
+		// We only need to refresh local side if the operation got handled
+		// externally, the internal handlers do this for us already
 		m_pState->RefreshLocal();
+	}
 }
 
 #ifndef __WXMSW__
@@ -913,7 +1022,7 @@ bool CLocalTreeView::CreateRoot()
 		name = _("Desktop");
 		iconIndex = openIconIndex = -1;
 	}
-	
+
 	m_desktop = AddRoot(name, iconIndex, openIconIndex);
 
 	name = GetSpecialFolder(CSIDL_PERSONAL, iconIndex, openIconIndex);
@@ -922,7 +1031,7 @@ bool CLocalTreeView::CreateRoot()
 		name = _("My Documents");
 		iconIndex = openIconIndex = -1;
 	}
-	
+
 	m_documents = AppendItem(m_desktop, name, iconIndex, openIconIndex);
 
 
@@ -932,7 +1041,7 @@ bool CLocalTreeView::CreateRoot()
 		name = _("My Computer");
 		iconIndex = openIconIndex = -1;
 	}
-	
+
 	m_drives = AppendItem(m_desktop, name, iconIndex, openIconIndex);
 
 	DisplayDrives(m_drives);
@@ -950,8 +1059,11 @@ void CLocalTreeView::OnVolumesEnumerated(wxCommandEvent& event)
 	std::list<CVolumeDescriptionEnumeratorThread::t_VolumeInfo> volumeInfo;
 	volumeInfo = m_pVolumeEnumeratorThread->GetVolumes();
 
-	delete m_pVolumeEnumeratorThread;
-	m_pVolumeEnumeratorThread = 0;
+	if (event.GetEventType() == fzEVT_VOLUMESENUMERATED)
+	{
+		delete m_pVolumeEnumeratorThread;
+		m_pVolumeEnumeratorThread = 0;
+	}
 
 	for (std::list<CVolumeDescriptionEnumeratorThread::t_VolumeInfo>::const_iterator iter = volumeInfo.begin(); iter != volumeInfo.end(); iter++)
 	{
@@ -972,54 +1084,410 @@ void CLocalTreeView::OnVolumesEnumerated(wxCommandEvent& event)
 	}
 }
 
-DEFINE_EVENT_TYPE(fzEVT_VOLUMESENUMERATED)
+#endif
 
-CVolumeDescriptionEnumeratorThread::CVolumeDescriptionEnumeratorThread(wxEvtHandler* pEvtHandler)
-	: wxThreadEx(wxTHREAD_JOINABLE),
-	  m_pEvtHandler(pEvtHandler)
+void CLocalTreeView::OnContextMenu(wxTreeEvent& event)
 {
-	m_failure = false;
-	m_stop = false;
-	m_running = true;
+	m_contextMenuItem = event.GetItem();
+	if (!m_contextMenuItem.IsOk())
+		return;
 
-	if (Create() != wxTHREAD_NO_ERROR)
+	wxMenu* pMenu = wxXmlResource::Get()->LoadMenu(_T("ID_MENU_LOCALTREE"));
+	if (!pMenu)
+		return;
+
+	const CLocalPath path(GetDirFromItem(m_contextMenuItem));
+
+	const bool hasParent = path.HasParent();
+	const bool writeable = path.IsWriteable();
+
+	const bool remoteConnected = m_pState->IsRemoteConnected() && !m_pState->GetRemotePath().IsEmpty();
+
+	pMenu->Enable(XRCID("ID_UPLOAD"), hasParent && remoteConnected);
+	pMenu->Enable(XRCID("ID_ADDTOQUEUE"), hasParent && remoteConnected);
+	pMenu->Enable(XRCID("ID_MKDIR"), writeable);
+	pMenu->Enable(XRCID("ID_DELETE"), writeable && hasParent);
+	pMenu->Enable(XRCID("ID_RENAME"), writeable && hasParent);
+
+	PopupMenu(pMenu);
+	delete pMenu;
+}
+
+void CLocalTreeView::OnMenuUpload(wxCommandEvent& event)
+{
+	if (!m_contextMenuItem.IsOk())
+		return;
+
+	wxString path = GetDirFromItem(m_contextMenuItem);
+
+	if (!CLocalPath(path).HasParent())
+		return;
+
+	if (!m_pState->IsRemoteConnected())
+		return;
+
+	const CServer server = *m_pState->GetServer();
+	CServerPath remotePath = m_pState->GetRemotePath();
+	if (remotePath.IsEmpty())
+		return;
+
+	if (!remotePath.ChangePath(GetItemText(m_contextMenuItem)))
+		return;
+
+	if (path.Last() == wxFileName::GetPathSeparator())
+		path.RemoveLast();
+	m_pQueueView->QueueFolder(event.GetId() == XRCID("ID_ADDTOQUEUE"), false, path, remotePath, server);
+}
+
+void CLocalTreeView::OnMenuMkdir(wxCommandEvent& event)
+{
+	if (!m_contextMenuItem.IsOk())
+		return;
+
+	wxString path = GetDirFromItem(m_contextMenuItem);
+	if (path.Last() != wxFileName::GetPathSeparator())
+		path += wxFileName::GetPathSeparator();
+
+	if (!CLocalPath(path).IsWriteable())
 	{
-		m_running = false;
-		m_failure = true;
+		wxBell();
+		return;
 	}
-	Run();
-}
 
-CVolumeDescriptionEnumeratorThread::~CVolumeDescriptionEnumeratorThread()
-{
-	m_stop = true;
-	Wait();
+	CInputDialog dlg;
+	if (!dlg.Create(this, _("Create directory"), _("Please enter the name of the directory which should be created:")))
+		return;
 
-	for (std::list<t_VolumeInfoInternal>::const_iterator iter = m_volumeInfo.begin(); iter != m_volumeInfo.end(); iter++)
+	wxString newName = _("New directory");
+	dlg.SetValue(path + newName);
+	dlg.SelectText(path.Len(), path.Len() + newName.Len());
+
+	if (dlg.ShowModal() != wxID_OK)
+		return;
+
+	wxFileName fn(dlg.GetValue(), _T(""));
+	if (!fn.Normalize(wxPATH_NORM_ALL, path))
 	{
-		delete [] iter->pVolume;
-		delete [] iter->pVolumeName;
+		wxBell();
+		return;
 	}
-	m_volumeInfo.clear();
+
+	bool res;
+	{
+		wxLogNull log;
+		res = fn.Mkdir(fn.GetPath(), 0777, wxPATH_MKDIR_FULL);
+	}
+
+	if (!res)
+		wxBell();
+
+	Refresh();
+	m_pState->RefreshLocal();
 }
 
-wxThread::ExitCode CVolumeDescriptionEnumeratorThread::Entry()
+void CLocalTreeView::OnMenuRename(wxCommandEvent& event)
 {
-	if (!GetDrives())
-		m_failure = true;
+	if (!m_contextMenuItem.IsOk())
+		return;
 
-	m_running = false;
+#ifdef __WXMSW__
+	if (m_contextMenuItem == m_desktop || m_contextMenuItem == m_documents)
+	{
+		wxBell();
+		return;
+	}
+#endif
 
-	wxCommandEvent evt(fzEVT_VOLUMESENUMERATED);
-	m_pEvtHandler->AddPendingEvent(evt);
+	CLocalPath path(GetDirFromItem(m_contextMenuItem));
+	if (!path.HasParent() || !path.IsWriteable())
+	{
+		wxBell();
+		return;
+	}
 
-	return 0;
+	EditLabel(m_contextMenuItem);
 }
 
-bool CVolumeDescriptionEnumeratorThread::GetDrives()
+void CLocalTreeView::OnMenuDelete(wxCommandEvent& event)
+{
+	if (!m_contextMenuItem.IsOk())
+		return;
+
+	wxString path = GetDirFromItem(m_contextMenuItem);
+
+	CLocalPath local_path(path);
+	if (!local_path.HasParent() || !local_path.IsWriteable())
+		return;
+
+	if (!CLocalFileSystem::RecursiveDelete(path, this))
+		wxGetApp().DisplayEncodingWarning();
+
+	wxTreeItemId item = GetSelection();
+	while (item && item != m_contextMenuItem)
+		item = GetItemParent(item);
+
+	if (!item)
+	{
+		if (GetItemParent(m_contextMenuItem) == GetSelection())
+			m_pState->RefreshLocal();
+		else
+			Refresh();
+		return;
+	}
+
+	if (path.Last() == wxFileName::GetPathSeparator())
+		path.RemoveLast();
+	int pos = path.Find(wxFileName::GetPathSeparator(), true);
+	if (pos < 1)
+		path = _T("/");
+	else
+		path = path.Left(pos);
+
+	m_pState->SetLocalDir(path);
+	Refresh();
+}
+
+void CLocalTreeView::OnBeginLabelEdit(wxTreeEvent& event)
+{
+	wxTreeItemId item = event.GetItem();
+
+#ifdef __WXMSW__
+	if (item == m_desktop || item == m_documents)
+	{
+		wxBell();
+		event.Veto();
+		return;
+	}
+#endif
+
+	CLocalPath path(GetDirFromItem(item));
+
+	if (!path.HasParent() || !path.IsWriteable())
+	{
+		wxBell();
+		event.Veto();
+		return;
+	}
+}
+
+// Defined in LocalListView.cpp
+extern bool Rename(wxWindow* pWnd, wxString dir, wxString from, wxString to);
+
+void CLocalTreeView::OnEndLabelEdit(wxTreeEvent& event)
+{
+	if (event.IsEditCancelled())
+	{
+		event.Veto();
+		return;
+	}
+
+	wxTreeItemId item = event.GetItem();
+
+#ifdef __WXMSW__
+	if (item == m_desktop || item == m_documents)
+	{
+		wxBell();
+		event.Veto();
+		return;
+	}
+#endif
+
+	wxString path = GetDirFromItem(item);
+
+	CLocalPath local_path(path);
+	if (!local_path.HasParent() || !local_path.IsWriteable())
+	{
+		wxBell();
+		event.Veto();
+		return;
+	}
+
+	if (path.Last() == wxFileName::GetPathSeparator())
+		path.RemoveLast();
+
+	int pos = path.Find(wxFileName::GetPathSeparator(), true);
+	wxASSERT(pos != -1);
+
+	wxString parent = path.Left(pos + 1);
+
+	const wxString& oldName = GetItemText(item);
+	const wxString& newName = event.GetLabel();
+	if (newName == _T(""))
+	{
+		wxBell();
+		event.Veto();
+		return;
+	}
+
+	wxASSERT(parent + oldName == path);
+
+	if (oldName == newName)
+		return;
+
+	if (!Rename(this, parent, oldName, newName))
+	{
+		event.Veto();
+		return;
+	}
+
+	// We may call SetLocalDir, item might be deleted by it, so
+	// if we don't rename item now and veto the event, wxWidgets
+	// might access deleted item.
+	event.Veto();
+	SetItemText(item, newName);
+
+	wxTreeItemId currentSel = GetSelection();
+	if (currentSel == wxTreeItemId())
+	{
+		Refresh();
+		return;
+	}
+
+	if (item == currentSel)
+	{
+		m_pState->SetLocalDir(parent + newName);
+		return;
+	}
+
+	wxString sub;
+
+	wxTreeItemId tmp = currentSel;
+	while (tmp != GetRootItem() && tmp != item)
+	{
+		sub = wxFileName::GetPathSeparator() + GetItemText(tmp) + sub;
+		tmp = GetItemParent(tmp);
+	}
+
+	if (tmp == GetRootItem())
+	{
+		// Rename unrelated to current selection
+		return;
+	}
+
+	// Current selection below renamed item
+	m_pState->SetLocalDir(parent + newName + sub);
+}
+
+void CLocalTreeView::OnChar(wxKeyEvent& event)
+{
+	m_contextMenuItem = GetSelection();
+
+	wxCommandEvent cmdEvt;
+	if (event.GetKeyCode() == WXK_F2)
+		OnMenuRename(cmdEvt);
+	else if (event.GetKeyCode() == WXK_DELETE)
+		OnMenuDelete(cmdEvt);
+	else
+		event.Skip();
+}
+
+bool CLocalTreeView::CheckSubdirStatus(wxTreeItemId& item, const wxString& path)
+{
+	wxTreeItemIdValue value;
+	wxTreeItemId child = GetFirstChild(item, value);
+
+	static const wxLongLong size(-1);
+
+#ifdef __WXMAC__
+	// By default, OS X has a list of servers mounted into /net,
+	// listing that directory is slow.
+	if (GetItemParent(item) == GetRootItem() && (path == _T("/net") || path == _T("/net/")))
+	{
+			CFilterManager filter;
+
+			const int attributes = S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
+			if (!filter.FilenameFiltered(_T("localhost"), path, true, size, true, attributes))
+			{
+				if (!child)
+					AppendItem(item, _T(""));
+				return true;
+			}
+	}
+#endif
+
+	if (child)
+	{
+		if (GetItemText(child) != _T(""))
+			return false;
+
+		CTreeItemData* pData = (CTreeItemData*)GetItemData(child);
+		if (pData)
+		{
+			bool wasLink;
+			int attributes;
+			enum CLocalFileSystem::local_fileType type;
+			if (path.Last() == CLocalFileSystem::path_separator)
+				type = CLocalFileSystem::GetFileInfo(path + pData->m_known_subdir, wasLink, 0, 0, &attributes);
+			else
+				type = CLocalFileSystem::GetFileInfo(path + CLocalFileSystem::path_separator + pData->m_known_subdir, wasLink, 0, 0, &attributes);
+			if (type == CLocalFileSystem::dir)
+			{
+				CFilterManager filter;
+				if (!filter.FilenameFiltered(pData->m_known_subdir, path, true, size, true, attributes))
+					return true;
+			}
+		}
+	}
+
+	wxString sub = HasSubdir(path);
+	if (sub != _T(""))
+	{
+		wxTreeItemId subItem = AppendItem(item, _T(""));
+		SetItemData(subItem, new CTreeItemData(sub));
+	}
+	else if (child)
+		Delete(child);
+
+	return true;
+}
+
+#ifdef __WXMSW__
+void CLocalTreeView::OnDevicechange(WPARAM wParam, LPARAM lParam)
+{
+	if (!m_drives)
+		return;
+
+	if (wParam == DBT_DEVICEARRIVAL || wParam == DBT_DEVICEREMOVECOMPLETE)
+	{
+		DEV_BROADCAST_HDR* pHdr = (DEV_BROADCAST_HDR*)lParam;
+		if (pHdr->dbch_devicetype != DBT_DEVTYP_VOLUME)
+			return;
+
+		// Added or removed volume
+		DEV_BROADCAST_VOLUME* pVolume = (DEV_BROADCAST_VOLUME*)lParam;
+
+		wxChar drive = 'A';
+		int mask = 1;
+		while (drive <= 'Z')
+		{
+			if (pVolume->dbcv_unitmask & mask)
+			{
+				if (wParam == DBT_DEVICEARRIVAL)
+					AddDrive(drive);
+				else
+				{
+					RemoveDrive(drive);
+
+					if (pVolume->dbcv_flags & DBTF_MEDIA)
+					{
+						// E.g. disk removed from CD-ROM drive, need to keep the drive letter
+						AddDrive(drive);
+					}
+				}
+			}
+			drive++;
+			mask <<= 1;
+		}
+
+		if (GetSelection() == m_drives && m_pState)
+			m_pState->RefreshLocal();
+	}
+}
+
+wxTreeItemId CLocalTreeView::AddDrive(wxChar drive)
 {
-	long drivesToHide = 0;
 	// Adhere to the NODRIVES group policy
+	long drivesToHide = 0;
 	wxRegKey key(_T("HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer"));
 	if (key.Exists())
 	{
@@ -1027,101 +1495,76 @@ bool CVolumeDescriptionEnumeratorThread::GetDrives()
 			drivesToHide = 0;
 	}
 
-	int len = GetLogicalDriveStrings(0, 0);
-	if (!len)
-		return false;
+	int bit = 0;
+	if (drive >= 'A' && drive <= 'Z')
+		bit = 1 << (drive - 'A');
+	else if (drive >= 'a' && drive <= 'z')
+		bit = 1 << (drive - 'a');
+	if (drivesToHide & bit)
+		return wxTreeItemId();
 
-	wxChar* drives = new wxChar[len + 1];
+	wxString driveName = drive;
+	driveName += _T(":");
 
-	if (!GetLogicalDriveStrings(len, drives))
+	// Get the label of the drive
+	wxChar volumeName[501];
+	int oldErrorMode = SetErrorMode(SEM_FAILCRITICALERRORS);
+	BOOL res = GetVolumeInformation(driveName + _T("\\"), volumeName, 500, 0, 0, 0, 0, 0);
+	SetErrorMode(oldErrorMode);
+
+	wxString itemLabel = driveName;
+	if (res && volumeName[0])
 	{
-		delete [] drives;
-		return false;
+		itemLabel += _T(" (");
+		itemLabel += volumeName;
+		itemLabel += _T(")");
 	}
 
-	const wxChar* pDrive = drives;
-	while (*pDrive)
+	wxTreeItemIdValue value;
+	wxTreeItemId driveItem = GetFirstChild(m_drives, value);
+	while (driveItem)
 	{
-		if (m_stop)
-		{
-			delete [] drives;
-			return false;
-		}
+		if (!GetItemText(driveItem).Left(2).CmpNoCase(driveName))
+			break;
 
-		// Check if drive should be hidden by default
-		if (pDrive[0] != 0 && pDrive[1] == ':')
-		{
-			int bit = -1;
-			char letter = pDrive[0];
-			if (letter >= 'A' && letter <= 'Z')
-				bit = 1 << (letter - 'A');
-			if (letter >= 'a' && letter <= 'z')
-				bit = 1 << (letter - 'a');
+		driveItem = GetNextSibling(driveItem);
+	}
+	if (driveItem)
+	{
+		SetItemText(driveItem, itemLabel);
+		int icon = GetIconIndex(dir, driveName + _T("\\"));
+		SetItemImage(driveItem, icon, wxTreeItemIcon_Normal);
+		SetItemImage(driveItem, icon, wxTreeItemIcon_Selected);
+		SetItemImage(driveItem, icon, wxTreeItemIcon_Expanded);
+		SetItemImage(driveItem, icon, wxTreeItemIcon_SelectedExpanded);
 
-			if (bit != -1 && drivesToHide & bit)
-			{
-				pDrive += wxStrlen(pDrive) + 1;
-				continue;
-			}
-		}
-
-		int len = wxStrlen(pDrive);
-		wxChar* pVolume = new wxChar[len + 1];
-		wxStrcpy(pVolume, pDrive);
-		if (pVolume[len - 1] == '\\')
-			pVolume[len - 1] = 0;
-		if (!*pVolume)
-		{
-			delete [] pVolume;
-			pDrive += len + 1;
-			continue;
-		}
-
-		// Get the label of the drive
-		wxChar* pVolumeName = new wxChar[501];
-		int oldErrorMode = SetErrorMode(SEM_FAILCRITICALERRORS);
-		BOOL res = GetVolumeInformation(pDrive, pVolumeName, 500, 0, 0, 0, 0, 0);
-		SetErrorMode(oldErrorMode);
-		if (res && pVolumeName[0])
-		{
-			t_VolumeInfoInternal volumeInfo;
-			volumeInfo.pVolume = pVolume;
-			volumeInfo.pVolumeName = pVolumeName;
-			m_volumeInfo.push_back(volumeInfo);
-		}
-		else
-		{
-			delete [] pVolumeName;
-			delete [] pVolume;
-		}
-
-		pDrive += wxStrlen(pDrive) + 1;
+		return driveItem;
 	}
 
-	delete [] drives;
+	wxTreeItemId item = AppendItem(m_drives, itemLabel, GetIconIndex(dir, driveName + _T("\\")));
+	AppendItem(item, _T(""));
+	SortChildren(m_drives);
 
-	return true;
+	return item;
 }
 
-std::list<CVolumeDescriptionEnumeratorThread::t_VolumeInfo> CVolumeDescriptionEnumeratorThread::GetVolumes()
+void CLocalTreeView::RemoveDrive(wxChar drive)
 {
-	std::list<t_VolumeInfo> volumeInfo;
-
-	if (m_running)
-		return volumeInfo;
-
-	for (std::list<t_VolumeInfoInternal>::const_iterator iter = m_volumeInfo.begin(); iter != m_volumeInfo.end(); iter++)
+	wxString driveName = drive;
+	driveName += _T(":");
+	wxTreeItemIdValue value;
+	wxTreeItemId driveItem = GetFirstChild(m_drives, value);
+	while (driveItem)
 	{
-		t_VolumeInfo info;
-		info.volume = iter->pVolume;
-		delete [] iter->pVolume;
-		info.volumeName = iter->pVolumeName;
-		delete [] iter->pVolumeName;
-		volumeInfo.push_back(info);
-	}
-	m_volumeInfo.clear();
+		if (!GetItemText(driveItem).Left(2).CmpNoCase(driveName))
+			break;
 
-	return volumeInfo;
+		driveItem = GetNextSibling(driveItem);
+	}
+	if (!driveItem)
+		return;
+
+	Delete(driveItem);
 }
 
 #endif
