@@ -5,7 +5,11 @@
 #include "ipcmutex.h"
 #include "xmlfunctions.h"
 #include "buildinfo.h"
+#include "Options.h"
 
+bool CWrapEngine::m_use_cache = true;
+
+#define WRAPDEBUG 0
 #if wxUSE_UNICODE
 // Chinese equivalents to ".", "," and ":"
 static const wxChar noWrapChars_Chinese[] = { '.', ',', ':', 0x3002, 0xFF0C, 0xFF1A, 0};
@@ -73,7 +77,7 @@ bool CWrapEngine::WrapTextChinese(wxWindow* parent, wxString &text, unsigned lon
 			if (iter == m_charWidths.end())
 			{
 				// Get width of all individual characters, record width of the current line
-				parent->GetTextExtent(*p, &width, &height);
+				parent->GetTextExtent(*p, &width, &height, 0, 0, &m_font);
 				if ((unsigned int)width > maxLength)
 					return false;
 				m_charWidths[*p] = width;
@@ -163,7 +167,7 @@ bool CWrapEngine::WrapTextChinese(wxWindow* parent, wxString &text, unsigned lon
 			piece = temp.Left(pos);
 			temp = temp.Mid(pos + 1);
 		}
-		parent->GetTextExtent(piece, &width, &height);
+		parent->GetTextExtent(piece, &width, &height, 0, 0, &m_font);
 		wxASSERT(width <= (int)maxLength);
 	}
 #endif
@@ -195,6 +199,9 @@ bool CWrapEngine::WrapText(wxWindow* parent, wxString& text, unsigned long maxLe
 	For this kind of languages, a different wrapping algorithm is used.
     */
 
+	if (!m_font.IsOk())
+		m_font = parent->GetFont();
+
 	if (m_wrapOnEveryChar)
 	{
 #ifdef __WXDEBUG__
@@ -213,106 +220,54 @@ bool CWrapEngine::WrapText(wxWindow* parent, wxString& text, unsigned long maxLe
 	int width = 0, height = 0;
 
 	int spaceWidth = 0;
-	parent->GetTextExtent(_T(" "), &spaceWidth, &height);
+	parent->GetTextExtent(_T(" "), &spaceWidth, &height, 0, 0, &m_font);
 
 	int strLen = text.Length();
 	int wrapAfter = -1;
 	int start = 0;
 	unsigned int lineLength = 0;
+
+	bool url = false;
 	for (int i = 0; i <= strLen; i++)
 	{
+		if ((text[i] == ':' && text[i + 1] == '/' && text[i + 2] == '/') || // absolute
+			(text[i] == '/' && (!i || text[i - 1] == ' '))) // relative
+			url = true;
 		if (text[i] != ' ' && text[i] != 0)
+		{
+			// If url, wrap on slashes and ampersands, but not first slash of something://
+			if (!url ||
+				 ((text[i] != '/' || text[i + 1] == '/') && (text[i] != '&' || text[i + 1] == '&') && text[i] != '?'))
 			continue;
+		}
 
 		wxString segment;
 		if (wrapAfter == -1)
 		{
-			segment = text.Mid(start, i - start);
+			if (text[i] == '/' || text[i] == '?' || text[i] == '&')
+				segment = text.Mid(start, i - start + 1);
+			else
+				segment = text.Mid(start, i - start);
 			wrapAfter = i;
 		}
 		else
-			segment = text.Mid(wrapAfter + 1, i - wrapAfter - 1);
+		{
+			if (text[i] == '/' || text[i] == '?' || text[i] == '&')
+				segment = text.Mid(wrapAfter + 1, i - wrapAfter);
+			else
+				segment = text.Mid(wrapAfter + 1, i - wrapAfter - 1);
+		}
 
 		segment = wxStripMenuCodes(segment);
-		parent->GetTextExtent(segment, &width, &height);
-
-		if ((unsigned int)width > maxLength)
-		{
-			// Something quite long. Perhaps it's a URL we can wrap at slashes?
-
-			while (segment != _T(""))
-			{
-				// Find longest possible subsegment
-				// which can be appended to current line, if there is a current line
-				unsigned int j;
-				int best = -1;
-
-				for (j = 1; j < segment.Len() - 1; j++)
-				{
-					if (segment[j] == '/' && segment[j + 1] != '/')
-					{
-						wxString left = segment.Left(j + 1);
-						int lwidth;
-						parent->GetTextExtent(left, &lwidth, &height);
-						if ((unsigned)lwidth <= maxLength)
-						{
-							if (lineLength && lineLength + spaceWidth + lwidth > maxLength)
-							{
-								if (best == -1)
-								{
-									best = j;
-									width = lwidth;
-								}
-
-								break;
-							}
-
-							best = j;
-							width = lwidth;
-						}
-						else
-							break;
-					}
-				}
-				if (best == -1)
-				{
-					// No suitable slash found
-					return false;
-				}
-
-				if (wrappedText != _T(""))
-					wrappedText += _T("\n");
-
-				if (lineLength)
-					wrappedText += text.Mid(start, wrapAfter - start);
-
-				if (lineLength && lineLength + spaceWidth + width > maxLength)
-					wrappedText += _T("\n");
-				else
-					wrappedText += _T(" ");
-				lineLength = 0;
-
-				wrappedText += segment.Left(best + 1);
-				segment = segment.Mid(best + 1);
-
-				parent->GetTextExtent(wrappedText, &width, &height);
-				parent->GetTextExtent(segment, &width, &height);
-				if ((unsigned)width <= maxLength)
-				{
-					break;
-				}
-			}
-
-			start = i - segment.Len();
-			wrapAfter = i;
-			lineLength = 0;
-		}
+		parent->GetTextExtent(segment, &width, &height, 0, 0, &m_font);
 
 		if (lineLength + spaceWidth + width > maxLength)
 		{
 			if (wrappedText != _T(""))
 				wrappedText += _T("\n");
 			wrappedText += text.Mid(start, wrapAfter - start);
+			if (text[wrapAfter] != ' ' && text[wrapAfter] != '\0')
+				wrappedText += text[wrapAfter];
 			if (width + spaceWidth >= (int)maxLength)
 			{
 				if (wrappedText != _T(""))
@@ -335,7 +290,7 @@ bool CWrapEngine::WrapText(wxWindow* parent, wxString& text, unsigned long maxLe
 			if (wrappedText != _T(""))
 				wrappedText += _T("\n");
 			wrappedText += text.Mid(start, i - start);
-			if (text[i] != ' ')
+			if (text[i] != ' ' && text[i] != '\0')
 				wrappedText += text[i];
 			start = i + 1;
 			wrapAfter = -1;
@@ -348,6 +303,9 @@ bool CWrapEngine::WrapText(wxWindow* parent, wxString& text, unsigned long maxLe
 			lineLength += width;
 			wrapAfter = i;
 		}
+
+		if (text[i] == ' ')
+			url = false;
 	}
 	if (start < strLen)
 	{
@@ -376,12 +334,28 @@ bool CWrapEngine::WrapText(wxWindow* parent, int id, unsigned long maxLength)
 	return true;
 }
 
-bool CWrapEngine::WrapRecursive(wxWindow* wnd, wxSizer* sizer, int max)
+#if WRAPDEBUG >= 3
+	#define plvl { for (int i = 0; i < level; i++) printf(" "); }
+#endif
+
+int CWrapEngine::WrapRecursive(wxWindow* wnd, wxSizer* sizer, int max)
 {
 	// This function auto-wraps static texts.
+	
+#if WRAPDEBUG >= 3
+	static int level = 1;
+    plvl printf("Enter\n");
+#endif
 
 	if (max <= 0)
-		return false;
+	{
+#if WRAPDEBUG >= 3
+		plvl printf("Leave: max <= 0\n");
+#endif
+		return wrap_failed;
+	}
+
+	int result = 0;
 
 	for (unsigned int i = 0; i < sizer->GetChildren().GetCount(); i++)
 	{
@@ -405,7 +379,7 @@ bool CWrapEngine::WrapRecursive(wxWindow* wnd, wxSizer* sizer, int max)
 
 		if (min.GetWidth() + item->GetPosition().x + lborder + rborder <= max)
 		    continue;
-
+			
 		wxWindow* window;
 		wxSizer* subSizer = 0;
 		if ((window = item->GetWindow()))
@@ -413,14 +387,25 @@ bool CWrapEngine::WrapRecursive(wxWindow* wnd, wxSizer* sizer, int max)
 			wxStaticText* text = wxDynamicCast(window, wxStaticText);
 			if (text)
 			{
-				if (max - rect.GetLeft() - rborder - 2 <= 0)
+#ifdef __WXMAC__
+				const int offset = 3;
+#else
+				const int offset = 2;
+#endif
+				if (max - rect.GetLeft() - rborder - offset <= 0)
 					continue;
-
+					
 				wxString str = text->GetLabel();
-				if (!WrapText(text, str, max - wxMax(0, rect.GetLeft()) - rborder - 2))
-					return false;
+				if (!WrapText(text, str, max - wxMax(0, rect.GetLeft()) - rborder - offset))
+				{
+#if WRAPDEBUG >= 3
+					plvl printf("Leave: WrapText failed\n");
+#endif				
+					return wrap_failed;
+				}
 				text->SetLabel(str);
 
+				result |= wrap_didwrap;
 				continue;
 			}
 
@@ -439,13 +424,20 @@ bool CWrapEngine::WrapRecursive(wxWindow* wnd, wxSizer* sizer, int max)
 					wxNotebookPage* page = book->GetPage(i);
 					wxRect pageRect = page->GetRect();
 					int pageMax = max - rect.GetLeft() - pageRect.GetLeft() - rborder - rect.GetWidth() + maxPageWidth;
-					if (!WrapRecursive(wnd, page->GetSizer(), pageMax))
-						return false;
+
+					result |= WrapRecursive(wnd, page->GetSizer(), pageMax);
+					if (result & wrap_failed)
+					{
+#if WRAPDEBUG >= 3
+						plvl printf("Leave: WrapRecursive on notebook page failed\n");
+#endif
+						return result;
+					}
 				}
 				continue;
 			}
 		}
-		if ((subSizer = item->GetSizer()))
+		else if ((subSizer = item->GetSizer()))
 		{
 			int subBorder = 0;
 
@@ -458,19 +450,61 @@ bool CWrapEngine::WrapRecursive(wxWindow* wnd, wxSizer* sizer, int max)
 				subBorder += other;
 			}
 
-			if (!WrapRecursive(0, subSizer, max - rborder - subBorder))
-				return false;
+#if WRAPDEBUG >= 3
+			level++;
+#endif
+			result |= WrapRecursive(0, subSizer, max - rborder - subBorder);
+#if WRAPDEBUG >= 3
+			level--;
+#endif
+			if (result & wrap_failed)
+			{
+#if WRAPDEBUG >= 3
+				plvl printf("Leave: WrapRecursive on sizer failed\n");
+#endif
+				return result;
+			}
 		}
 	}
+	
+#if WRAPDEBUG >= 3
+	plvl printf("Leave: Success\n");
+#endif
 
-	return true;
+
+	return result;
 }
 
 bool CWrapEngine::WrapRecursive(wxWindow* wnd, double ratio, const char* name /*=""*/, wxSize canvas /*=wxSize()*/, wxSize minRequestedSize /*wxSize()*/)
 {
 	std::vector<wxWindow*> windows;
 	windows.push_back(wnd);
-	return WrapRecursive(windows, ratio, name, canvas, minRequestedSize);
+	return (WrapRecursive(windows, ratio, name, canvas, minRequestedSize) & wrap_failed) == 0;
+}
+
+void CWrapEngine::UnwrapRecursive_Wrapped(const std::list<int> &wrapped, std::vector<wxWindow*> &windows, bool remove_fitting /*=false*/)
+{
+	unsigned int i = 0;
+	for (std::list<int>::const_iterator iter = wrapped.begin();
+		iter != wrapped.end();
+		iter++)
+	{
+		if (!(*iter & wrap_didwrap))
+		{
+			if (!(*iter) && remove_fitting)
+			{
+				// Page didn't need to be wrapped with current wrap offset,
+				// remove it since desired with will only be larger in further wrappings.
+				windows.erase(windows.begin() + i);
+			}
+			continue;
+		}
+
+		UnwrapRecursive(windows[i], windows[i]->GetSizer());
+		windows[i]->GetSizer()->Layout();
+
+		i++;
+	}
 }
 
 bool CWrapEngine::WrapRecursive(std::vector<wxWindow*>& windows, double ratio, const char* name /*=""*/, wxSize canvas /*=wxSize()*/, wxSize minRequestedSize /*wxSize()*/)
@@ -485,11 +519,18 @@ bool CWrapEngine::WrapRecursive(std::vector<wxWindow*>& windows, double ratio, c
 				continue;
 
 			pSizer->Layout();
-#ifdef __WXDEBUG__
-			bool res =
+			
+#ifdef __WXMAC__
+			const int offset = 4;
+#else
+			const int offset = 0;
 #endif
-			WrapRecursive(*iter, pSizer, maxWidth);
-			wxASSERT(res);
+
+#ifdef __WXDEBUG__
+			int res =
+#endif
+			WrapRecursive(*iter, pSizer, maxWidth - offset);
+			wxASSERT(!(res & wrap_failed));
 			pSizer->Layout();
 			pSizer->Fit(*iter);
 			wxSize size = pSizer->GetMinSize();
@@ -528,39 +569,62 @@ bool CWrapEngine::WrapRecursive(std::vector<wxWindow*>& windows, double ratio, c
 
 	double bestRatioDiff = currentRatio - ratio;
 	int bestWidth = max;
+	
+#if WRAPDEBUG > 0	
+	printf("Target ratio: %f\n", (float)ratio);
+	printf("Canvas: % 4d % 4d\n", canvas.x, canvas.y);
+#endif
 
 	while (true)
 	{
+		std::list<int> didwrap;
+
 		wxSize size = minRequestedSize;
 		for (std::vector<wxWindow*>::iterator iter = windows.begin(); iter != windows.end(); iter++)
 		{
 			wxSizer* pSizer = (*iter)->GetSizer();
-			WrapRecursive(*iter, pSizer, desiredWidth);
-			pSizer->Layout();
+#ifdef __WXMAC__
+			const int offset = 4;
+#else
+			const int offset = 0;
+#endif
+			int res = WrapRecursive(*iter, pSizer, desiredWidth - offset);
+			didwrap.push_back(res);
+			if (res & wrap_didwrap)
+				pSizer->Layout();
 			size.IncTo(pSizer->GetMinSize());
+			if (res & wrap_failed)
+				break;
 		}
 
+#if WRAPDEBUG > 0
+	    printf("Current: % 4d % 4d   desiredWidth: %d\n", size.GetWidth(), size.GetHeight(), desiredWidth);
+#endif
 		if (size.GetWidth() > desiredWidth)
 		{
 			// Wrapping failed
+
+			UnwrapRecursive_Wrapped(didwrap, windows, true);
+
 			min = desiredWidth;
 			if (max - min < 5)
 				break;
+
 			desiredWidth = (min + max) / 2;
 
-			for (std::vector<wxWindow*>::iterator iter = windows.begin(); iter != windows.end(); iter++)
-			{
-				UnwrapRecursive(*iter, (*iter)->GetSizer());
-				(*iter)->GetSizer()->Layout();
-			}
 			continue;
 		}
 		actualWidth = size.GetWidth();
 
 		double newRatio = ((double)(size.GetWidth() + canvas.x) / (size.GetHeight() + canvas.y));
+#if WRAPDEBUG > 0		
+		printf("Ratio: %f\n", (float)newRatio);
+#endif		
 
 		if (newRatio < ratio)
 		{
+			UnwrapRecursive_Wrapped(didwrap, windows, true);
+
 			if (ratio - newRatio < bestRatioDiff)
 			{
 				bestRatioDiff = ratio - newRatio;
@@ -574,6 +638,7 @@ bool CWrapEngine::WrapRecursive(std::vector<wxWindow*>& windows, double ratio, c
 		}
 		else if (newRatio > ratio)
 		{
+			UnwrapRecursive_Wrapped(didwrap, windows);
 			if (newRatio - ratio < bestRatioDiff)
 			{
 				bestRatioDiff = newRatio - ratio;
@@ -586,6 +651,8 @@ bool CWrapEngine::WrapRecursive(std::vector<wxWindow*>& windows, double ratio, c
 		}
 		else
 		{
+			UnwrapRecursive_Wrapped(didwrap, windows);
+
 			bestRatioDiff = ratio - newRatio;
 			bestWidth = actualWidth;
 			break;
@@ -595,25 +662,23 @@ bool CWrapEngine::WrapRecursive(std::vector<wxWindow*>& windows, double ratio, c
 			break;
 		desiredWidth = (min + max) / 2;
 
-		for (std::vector<wxWindow*>::iterator iter = windows.begin(); iter != windows.end(); iter++)
-		{
-			UnwrapRecursive(*iter, (*iter)->GetSizer());
-			(*iter)->GetSizer()->Layout();
-		}
-
 		currentRatio = newRatio;
 	}
 	for (std::vector<wxWindow*>::iterator iter = windows.begin(); iter != windows.end(); iter++)
 	{
-		wxSizer* pSizer = (*iter)->GetSizer();
-		UnwrapRecursive(*iter, pSizer);
-		pSizer->Layout();
+		wxSizer *pSizer = (*iter)->GetSizer();
 
-		WrapRecursive(*iter, pSizer, bestWidth);
-		pSizer->Layout();
-		pSizer->Fit(*iter);
+		int res = WrapRecursive(*iter, pSizer, bestWidth);
+
+		if (res & wrap_didwrap)
+		{
+			pSizer->Layout();
+			pSizer->Fit(*iter);
+		}
+#ifdef __WXDEBUG__
 		size = pSizer->GetMinSize();
 		wxASSERT(size.x <= bestWidth);
+#endif
 	}
 
 	SetWidthToCache(name, bestWidth);
@@ -709,6 +774,9 @@ bool CWrapEngine::UnwrapRecursive(wxWindow* wnd, wxSizer* sizer)
 
 int CWrapEngine::GetWidthFromCache(const char* name)
 {
+	if (!m_use_cache)
+		return 0;
+
 	if (!name || !*name)
 		return 0;
 
@@ -729,9 +797,11 @@ int CWrapEngine::GetWidthFromCache(const char* name)
 		return 0;
 	}
 
-	int language = wxGetApp().GetCurrentLanguage();
+	wxString language = wxGetApp().GetCurrentLanguageCode();
+	if (language.empty())
+		language = _T("default");
 
-	TiXmlElement* pLanguage = FindElementWithAttribute(pElement, "Language", "id", language);
+	TiXmlElement* pLanguage = FindElementWithAttribute(pElement, "Language", "id", language.mb_str());
 	if (!pLanguage)
 	{
 		delete pDocument->GetDocument();
@@ -754,6 +824,9 @@ int CWrapEngine::GetWidthFromCache(const char* name)
 
 void CWrapEngine::SetWidthToCache(const char* name, int width)
 {
+	if (!m_use_cache)
+		return;
+
 	if (!name || !*name)
 		return;
 
@@ -774,9 +847,11 @@ void CWrapEngine::SetWidthToCache(const char* name, int width)
 		return;
 	}
 
-	int language = wxGetApp().GetCurrentLanguage();
+	wxString language = wxGetApp().GetCurrentLanguageCode();
+	if (language.empty())
+		language = _T("default");
 
-	TiXmlElement* pLanguage = FindElementWithAttribute(pElement, "Language", "id", language);
+	TiXmlElement* pLanguage = FindElementWithAttribute(pElement, "Language", "id", language.mb_str());
 	if (!pLanguage)
 	{
 		delete pDocument->GetDocument();
@@ -806,6 +881,36 @@ CWrapEngine::~CWrapEngine()
 {
 }
 
+static wxString GetLocaleFile(const wxString& localesDir, wxString name)
+{
+	if (wxFileName::FileExists(localesDir + name + _T("/filezilla.mo")))
+		return name;
+	if (wxFileName::FileExists(localesDir + name + _T("/LC_MESSAGES/filezilla.mo")))
+		return name + _T("/LC_MESSAGES");
+
+	size_t pos = name.Find('@');
+	if (pos > 0)
+	{
+		name = name.Left(pos);
+		if (wxFileName::FileExists(localesDir + name + _T("/filezilla.mo")))
+			return name;
+		if (wxFileName::FileExists(localesDir + name + _T("/LC_MESSAGES/filezilla.mo")))
+			return name + _T("/LC_MESSAGES");
+	}
+
+	pos = name.Find('_');
+	if (pos > 0)
+	{
+		name = name.Left(pos);
+		if (wxFileName::FileExists(localesDir + name + _T("/filezilla.mo")))
+			return name;
+		if (wxFileName::FileExists(localesDir + name + _T("/LC_MESSAGES/filezilla.mo")))
+			return name + _T("/LC_MESSAGES");
+	}
+
+	return _T("");
+}
+
 bool CWrapEngine::LoadCache()
 {
 	// We have to synchronize access to layout.xml so that multiple processed don't write
@@ -813,12 +918,13 @@ bool CWrapEngine::LoadCache()
 	CInterProcessMutex mutex(MUTEX_LAYOUT);
 
 	wxFileName file(wxGetApp().GetSettingsDir(), _T("layout.xml"));
-	TiXmlElement* pDocument = GetXmlFile(file);
+	CXmlFile xml(file);
+	TiXmlElement* pDocument = xml.Load();
 
 	if (!pDocument)
 	{
-		wxString msg = wxString::Format(_("Could not load \"%s\", please make sure the file is valid and can be accessed."), file.GetFullPath().c_str());
-		wxMessageBox(msg, _("Error loading xml file"), wxICON_ERROR);
+		m_use_cache = false;
+		wxMessageBox(xml.GetError(), _("Error loading xml file"), wxICON_ERROR);
 
 		return false;
 	}
@@ -885,38 +991,6 @@ bool CWrapEngine::LoadCache()
 		}
 	}
 
-	// Get static text font and measure sample text
-	wxFrame* pFrame = new wxFrame;
-	pFrame->Create(0, -1, _T("Title"), wxDefaultPosition, wxDefaultSize, wxFRAME_TOOL_WINDOW);
-	wxStaticText* pText = new wxStaticText(pFrame, -1, _T("foo"));
-
-	wxFont font = pText->GetFont();
-	wxString fontDesc = font.GetNativeFontInfoDesc();
-
-	TiXmlElement* pFontElement = pElement->FirstChildElement("Font");
-	if (!pFontElement)
-		pFontElement = pElement->InsertEndChild(TiXmlElement("Font"))->ToElement();
-
-	if (GetTextAttribute(pFontElement, "font") != fontDesc)
-	{
-		SetTextAttribute(pFontElement, "font", fontDesc);
-		cacheValid = false;
-	}
-
-	int width, height;
-	pText->GetTextExtent(_T("Just some test string we are measuring. If width or heigh differ from the recorded values, invalidate cache."), &width, &height);
-
-	if (GetAttributeInt(pFontElement, "width") != width ||
-		GetAttributeInt(pFontElement, "height") != height)
-	{
-		cacheValid = false;
-		SetAttributeInt(pFontElement, "width", width);
-		SetAttributeInt(pFontElement, "height", height);
-	}
-
-	pFrame->Destroy();
-
-
 	if (!cacheValid)
 	{
 		// Clear all languages
@@ -928,68 +1002,90 @@ bool CWrapEngine::LoadCache()
 		}
 	}
 
-	// Enumerate language files
-	// ------------------------
+	// Get current language
+	wxString language = wxGetApp().GetCurrentLanguageCode();
+	if (language == _T(""))
+		language = _T("default");
 
-	const wxLanguageInfo* pInfo = wxLocale::FindLanguageInfo(_T("en"));
-	if (pInfo)
+	TiXmlElement* languageElement = FindElementWithAttribute(pElement, "Language", "id", language.mb_str());
+	if (!languageElement)
 	{
-		TiXmlElement* languageElement = FindElementWithAttribute(pElement, "Language", "id", pInfo->Language);
-		if (!languageElement)
-		{
-			languageElement = pElement->InsertEndChild(TiXmlElement("Language"))->ToElement();
-			languageElement->SetAttribute("id", pInfo->Language);
-		}
+		languageElement = pElement->InsertEndChild(TiXmlElement("Language"))->ToElement();
+		languageElement->SetAttribute("id", language.mb_str());
 	}
 
+	// Get static text font and measure sample text
+	wxFrame* pFrame = new wxFrame;
+	pFrame->Create(0, -1, _T("Title"), wxDefaultPosition, wxDefaultSize, wxFRAME_TOOL_WINDOW);
+	wxStaticText* pText = new wxStaticText(pFrame, -1, _T("foo"));
 
-	wxString localesDir = wxGetApp().GetLocalesDir();
-	if (localesDir != _T("") && dir.Open(localesDir))
+	wxFont font = pText->GetFont();
+	wxString fontDesc = font.GetNativeFontInfoDesc();
+
+	TiXmlElement* pFontElement = languageElement->FirstChildElement("Font");
+	if (!pFontElement)
+		pFontElement = languageElement->InsertEndChild(TiXmlElement("Font"))->ToElement();
+
+	if (GetTextAttribute(pFontElement, "font") != fontDesc)
 	{
-		wxString locale;
-		for (bool found = dir.GetFirst(&locale); found; found = dir.GetNext(&locale))
-		{
-			if (!wxFileName::FileExists(localesDir + locale + _T("/filezilla.mo")))
-				continue;
-
-			wxString name;
-			const wxLanguageInfo* pInfo = wxLocale::FindLanguageInfo(locale);
-			if (!pInfo)
-				continue;
-
-			wxFileName fn(localesDir + locale + _T("/filezilla.mo"));
-			wxDateTime date = fn.GetModificationTime();
-			wxLongLong ticks = date.GetTicks();
-
-			TiXmlElement* languageElement = FindElementWithAttribute(pElement, "Language", "id", pInfo->Language);
-			if (!languageElement)
-			{
-				languageElement = pElement->InsertEndChild(TiXmlElement("Language"))->ToElement();
-				languageElement->SetAttribute("id", pInfo->Language);
-				languageElement->SetAttribute("date", ticks.ToString().mb_str());
-			}
-			else
-			{
-				const char* languageNodeDate = languageElement->Attribute("date");
-				if (!languageNodeDate || strcmp(languageNodeDate, ticks.ToString().mb_str()))
-				{
-					languageElement->SetAttribute("date", ticks.ToString().mb_str());
-					languageElement->Clear();
-				}
-			}
-		}
+		SetTextAttribute(pFontElement, "font", fontDesc);
+		cacheValid = false;
 	}
 
-	// Outdated cache entries are now purged
+	int width, height;
+	pText->GetTextExtent(_T("Just some test string we are measuring. If width or heigh differ from the recorded values, invalidate cache. 1234567890MMWWII"), &width, &height);
+
+	if (GetAttributeInt(pFontElement, "width") != width ||
+		GetAttributeInt(pFontElement, "height") != height)
+	{
+		cacheValid = false;
+		SetAttributeInt(pFontElement, "width", width);
+		SetAttributeInt(pFontElement, "height", height);
+	}
+
+	pFrame->Destroy();
+
+	// Get language file
+	const wxString& localesDir = wxGetApp().GetLocalesDir();
+	wxString name = GetLocaleFile(localesDir, language);
+
+	if (name != _T(""))
+	{
+		wxFileName fn(localesDir + name + _T("/filezilla.mo"));
+		wxDateTime date = fn.GetModificationTime();
+		wxLongLong ticks = date.GetTicks();
+
+		const char* languageNodeDate = languageElement->Attribute("date");
+		if (!languageNodeDate || strcmp(languageNodeDate, ticks.ToString().mb_str()))
+		{
+			languageElement->SetAttribute("date", ticks.ToString().mb_str());
+			cacheValid = false;
+		}
+	}
+	else
+		languageElement->SetAttribute("date", "");
+	if (!cacheValid)
+	{
+		TiXmlElement* dialog;
+		while ((dialog = languageElement->FirstChildElement("Dialog")))
+			languageElement->RemoveChild(dialog);
+	}
+
+	if (COptions::Get()->GetDefaultVal(DEFAULT_KIOSKMODE) == 2)
+	{
+		m_use_cache = cacheValid;
+		return true;
+	}
 
 	wxString error;
-	if (!SaveXmlFile(file, pDocument, &error))
+	if (!xml.Save(&error))
 	{
+		m_use_cache = false;
+
 		wxString msg = wxString::Format(_("Could not write \"%s\": %s"), file.GetFullPath().c_str(), error.c_str());
 		wxMessageBox(msg, _("Error writing xml file"), wxICON_ERROR);
 	}
 
-	delete pDocument->GetDocument();
 
 	return true;
 }
@@ -1007,7 +1103,7 @@ void CWrapEngine::ClearCache()
 
 void CWrapEngine::CheckLanguage()
 {
-	#if wxUSE_UNICODE
+#if wxUSE_UNICODE
 	// Just don't bother with wrapping on anything other than UCS-2
 	// FIXME: Use charset conversion routines to convert into UCS-2 and back into
 	//        local charset if not using unicode.
